from birria._birria import (
    cook,
    ingredient,
    ingredients,
    is_cooked,
    is_cooked_instance,
    is_cooked_class,
    serve,
    _match_opt_strings
)

__version__ = "0.2.1"

__all__ = [
    "cook",
    "ingredient",
    "ingredients",
    "is_cooked",
    "is_cooked_instance",
    "is_cooked_class",
    "serve",
    "_match_opt_strings"
]



import builtins
from itertools import chain
from pathlib import PurePath
import re
from shutil import get_terminal_size
import sys
from types import MappingProxyType, MemberDescriptorType
from typing import (
    Any,
    Callable,
    List,
    Iterable,
    Tuple,
    TypeVar,
    Type,
    IO,
    Iterator,
    Dict,
    Mapping,
    Union,
    Optional,
)


class _HAS_DEFAULT_FACTORY_CLASS:
    def __repr__(self) -> str:
        return "<factory>"


class _MISSING_TYPE:
    pass


MISSING = _MISSING_TYPE()
_HAS_DEFAULT_FACTORY = _HAS_DEFAULT_FACTORY_CLASS()
_EMPTY_METADATA: Mapping = MappingProxyType({})

# list will be treated by the parser as List[str] (items are parsed as strings)
_SUPPORTED_TYPES = (
    str,
    int,
    float,
    bool,
    List[str],
    List[int],
    List[float],
    list,
    List,
)
_SUPPORTED_LIST_TYPES = (list, List[str], List[int], List[float], List)
_LIST_CASTS: Dict[Any, type] = {
    list: str,
    List: str,
    List[str]: str,
    List[int]: int,
    List[float]: float,
}
_TYPE_STR: Dict[Any, str] = {
    str: "string",
    int: "integer",
    float: "decimal",
    bool: "flag",
    list: "string(s)",
    List: "string(s)",
    List[str]: "string(s)",
    List[int]: "integer(s)",
    List[float]: "decimal(s)",
}
_HELP_OPT = ("help", "h")

# default prefixes used by the parser to match
# against option strings
_DEFAULT_PREFIXES = ["-"]

# raw strings for building regex
# to match against option strings
_ALLOWED_PREFIXES = {
    "-": r"\-",
    "+": r"\+",
    "/": r"\/",
}

CookedBirria = TypeVar("CookedBirria")
UserClass = TypeVar("UserClass")
OptionalType = Union[_MISSING_TYPE, Type, None]
CookedIngredient = Union[str, int, float, bool, List[int], List[float], List[str]]


# The name of an attribute on the class where we store the Field
# objects.  Also used to check if a class is a cooked class.
_FIELDS = "__arg_fields__"

# The name of an attribute on the class that stores the parameters to
# @dataclass.
_PARAMS = "__arg_params__"


class Ingredient:
    """Class that represents an ingredient.

    This class is used as a more elaborate declaration of an ingredient/argument.
    You should probably use ingredient() instead of calling this directly.
    """

    __slots__ = ("name", "type", "default", "default_factory", "help", "alias")

    def __init__(
        self,
        default: Any,
        default_factory: Callable[[], Any],
        help: Optional[str],
        alias: Optional[str],
    ):
        self.name: str = None  # type: ignore[assignment]
        self.type: type = None  # type: ignore[assignment]
        self.default = default
        self.default_factory = default_factory
        self.help = help
        self.alias = alias

    def __repr__(self) -> str:
        return (
            "Ingredient("
            f"name={self.name!r},"
            f"type={self.type!r},"
            f"default={self.default!r},"
            f"default_factory={self.default_factory!r},"
            f"help={self.help!r}"
            f"alias={self.alias!r}"
            ")"
        )

    # This is used to support the PEP 487 __set_name__ protocol in the
    # case where we're using a field that contains a descriptor as a
    # default value.  For details on __set_name__, see
    # https://www.python.org/dev/peps/pep-0487/#implementation-details.
    #
    # Note that in _process_class, this Field object is overwritten
    # with the default value, so the end result is a descriptor that
    # had __set_name__ called on it at the right time.
    def __set_name__(self, owner, name: str):
        func = getattr(type(self.default), "__set_name__", None)
        if func:
            # __set_name__ present, call it
            func(self.default, owner, name)


def ingredient(
    *, default=MISSING, default_factory=MISSING, help: str = None, alias: str = None
) -> Ingredient:
    """Instantiate and returns an Ingredient.

    This is a factory function to create Ingredient class instance.
    You probably should use this instead of using Ingredient directly.

    Args:
        default: Default value for the ingredient
        default_factory: Callable that takes no parameters and returns
            a value to be used as the default value.
        help: Short description of the ingredient that will be printed
            as part of the help string.
        alias: Alias for option strings. This is only applicable for
            optional ingredient.

    Returns:
        An Ingredient class instance.
    """
    if default is not MISSING and default_factory is not MISSING:
        raise ValueError("can't specify both default and default factory")
    return Ingredient(default, default_factory, help, alias)


def _create_fn(
    name: str,
    args: Iterable[str],
    body: Iterable[str],
    *,
    globals: dict = None,
    locals: dict = None,
    return_type: OptionalType = MISSING,
):
    if locals is None:
        locals = {}
    if "BUILTINS" not in locals:
        locals["BUILTINS"] = builtins

    return_annotation = ""
    if return_type is not MISSING:
        locals["_return_type"] = return_type
        return_annotation = "->_return_type"
    args_str = ",".join(args)
    body_str = "\n".join(f"  {b}" for b in body)

    # compute text for the entire function
    txt = f" def {name}({args_str}){return_annotation}:\n{body_str}"

    local_vars = ", ".join(locals.keys())
    txt = f"def __create_fn__({local_vars}):\n{txt}\n return {name}"
    ns: dict = {}
    exec(txt, globals, ns)
    return ns["__create_fn__"](**locals)


def _field_init(f: Ingredient, globals: dict, self_name: str):
    default_name = f"_dflt_{f.name}"
    if f.default_factory is not MISSING:
        globals[default_name] = f.default_factory
        value: Optional[str] = (
            f"{default_name}() "
            f"if {f.name} is _HAS_DEFAULT_FACTORY "
            f"else {f.name}"
        )
    else:
        if f.default is MISSING:
            # no default, simple assignment
            value = f.name
        elif f.default is not MISSING:
            globals[default_name] = f.default
            value = f.name

    return _field_assign(f.name, value, self_name)


def _field_assign(name: Optional[str], value: Optional[str], self_name: str) -> str:
    return f"{self_name}.{name}={value}"


def _init_param(f: Ingredient) -> str:
    # Return __init__ parameter string for this field
    # Ex: equivalent of "x:int=3"

    if f.default is MISSING and f.default_factory is MISSING:
        default = ""
    elif f.default is not MISSING:
        # There's a default, this will be the name that's used to look
        # it up.
        default = f"=_dflt_{f.name}"
    elif f.default_factory is not MISSING:
        # There's a factory function.  Set a marker.
        default = "=_HAS_DEFAULT_FACTORY"
    return f"{f.name}:_type_{f.name}{default}"


def _init_fn(fields: Iterable[Ingredient], self_name: str, globals: dict):
    locals: Dict[str, Any] = {f"_type_{f.name}": f.type for f in fields}
    locals.update({"MISSING": MISSING, "_HAS_DEFAULT_FACTORY": _HAS_DEFAULT_FACTORY})

    body_lines = [_field_init(f, locals, self_name) for f in fields]
    if not body_lines:
        body_lines = ["pass"]

    return _create_fn(
        "__init__",
        [self_name] + [_init_param(f) for f in fields],
        body_lines,
        locals=locals,
        globals=globals,
        return_type=None,
    )


def _repr_fn(fields: Iterable[Ingredient], globals: dict):
    return _create_fn(
        "__repr__",
        ("self",),
        [
            'return self.__class__.__qualname__ + f"('
            + ", ".join([f"{f.name}={{self.{f.name}!r}}" for f in fields])
            + ')"'
        ],
        globals=globals,
    )


def _eq_fn(self_tuple: str, other_tuple: str, globals: dict):
    return _create_fn(
        "__eq__",
        ("self", "other"),
        [
            "if other.__class__ is self.__class__:",
            f" return {self_tuple}=={other_tuple}",
            "return NotImplemented",
        ],
        globals=globals,
    )


def _set_new_attributes(cls, a_name: str, value: Any) -> bool:
    # Don't overwrite existing attribute
    # Return True if attribute already exists
    if a_name in cls.__dict__:
        return True
    setattr(cls, a_name, value)
    return False


def _get_ingredient(cls, a_name: str, a_type: type) -> Ingredient:
    default = getattr(cls, a_name, MISSING)
    f: Ingredient
    if isinstance(default, Ingredient):
        f = default
    else:
        # field in __slots__, so no default value
        if isinstance(default, MemberDescriptorType):
            default = MISSING
        f = ingredient(default=default)

    f.name = a_name
    f.type = a_type

    # disallow mutable defaults for known container types
    if isinstance(f.default, list):
        raise ValueError(
            f"mutable default {type(f.default)} for field "
            f"{f.name} is not allowed: use default_factory() for list"
        )
    if isinstance(f.default, (set, dict)):
        raise ValueError(
            f"mutable default {type(f.default)} for field "
            "dict and set are not supported as default values: use default_factory() for list"
        )
    return f


def _tuple_str(name: str, fields: List[Ingredient]) -> str:
    if not fields:
        return "()"
    # trailing commas for tuple with 1 item
    return f"({','.join([f'{name}.{f.name}' for f in fields])},)"


def _proc_class(cls: Type[UserClass]) -> Type[UserClass]:
    # print(f"Processing class {cls.__name__}")
    fields = {}

    if cls.__module__ in sys.modules:
        globals = sys.modules[cls.__module__].__dict__
    else:
        # Theoretically this can happen if someone writes
        # a custom string to cls.__module__.  In which case
        # such dataclass won't be fully introspectable
        # (w.r.t. typing.get_type_hints) but will still function
        # correctly.
        globals = {}

    base_non_default_fields = []
    base_default_fields = []
    for b in cls.__mro__[-1:0:-1]:
        # only include classes that use our decorator
        base_fields = getattr(b, _FIELDS, None)
        if base_fields:
            # print(f"Base field for {b.__name__}")
            # pprint(base_fields)
            for f in base_fields.values():
                if f.default is MISSING and f.default_factory is MISSING:
                    base_non_default_fields.append(f)
                else:
                    base_default_fields.append(f)

    # Annotations defined in this class and not the base classses.
    # if __annotations__ is not present, then this class adds no
    # new annotations. This is for fields added by this class.

    # Fields are from __annotations__, which is ordered. Defaults
    # are from class attributes.

    cls_annotations = cls.__dict__.get("__annotations__", {})
    cls_fields = []
    for fname, ftype in cls_annotations.items():
        if ftype not in _SUPPORTED_TYPES:
            raise TypeError(f"type {ftype} not supported")
        cls_fields.append(_get_ingredient(cls, fname, ftype))
    # pprint(cls_fields)

    # sort the fields into default and non-default
    non_default_fields = []
    default_fields = []

    for f in cls_fields:
        # if the class attribute exists and is a "Field"
        # replace it with the real default.
        attr = getattr(cls, f.name, MISSING)
        if isinstance(attr, Ingredient):
            if f.default is MISSING:
                # if no default set, delete the class attribute
                delattr(cls, f.name)
                if f.default_factory is MISSING:
                    non_default_fields.append(f)
                else:
                    default_fields.append(f)
            else:
                setattr(cls, f.name, f.default)
                default_fields.append(f)
        elif attr is MISSING:
            non_default_fields.append(f)
        else:
            default_fields.append(f)

    # Dict insertion is ordered, so default fields follow
    # non-default fields
    for f in chain(
        base_non_default_fields, non_default_fields, base_default_fields, default_fields
    ):
        # print(f"Inserting {f.name} in {cls.__name__}")
        fields[f.name] = f

    # raise TypeError for Field members with no annotations
    for name, val in cls.__dict__.items():
        if isinstance(val, Ingredient) and name not in cls_annotations:
            raise TypeError(f"{name!r} is a field but has no type annotations")

    if not fields:
        raise TypeError("No ingredients found in class or any base classes")

    # set all the fields to __arg_fields__
    # and mark this class as a cooked class
    # pprint(fields)
    setattr(cls, _FIELDS, fields)
    flds = list(fields.values())

    # attach __init__
    _set_new_attributes(
        cls,
        "__init__",
        _init_fn(flds, "__cooked_self__" if "self" in fields else "self", globals),
    )

    # attach __repr__
    _set_new_attributes(cls, "__repr__", _repr_fn(flds, globals))

    # attach __eq___
    # we want this for easy debugging and testing
    self_tuple = _tuple_str("self", flds)
    other_tuple = _tuple_str("other", flds)
    _set_new_attributes(cls, "__eq__", _eq_fn(self_tuple, other_tuple, globals=globals))

    return cls


# _cls should not be specified by keyword, hence the underscore,
# it's only used to detect if the decorator is called with params
# or not
def cook(_cls=None):
    """Decorator that takes a class and returns a CookedBirria class.

    Takes a class with typed variable annotations and a class based
    on the input class with __init__, __repr__, __eq__ dunder methods.
    This constructed class can then be used as a declaration of
    arguments for serve().
    """

    def wrap(cls):
        return _proc_class(cls)

    # called as @dataclass()
    if _cls is None:
        return wrap

    # called as @dataclass
    return wrap(_cls)


def ingredients(class_or_instance: object) -> Tuple[Ingredient, ...]:
    try:
        fields = getattr(class_or_instance, _FIELDS)
    except AttributeError:
        raise TypeError("must be called with a birria class type or instance")
    return tuple(fields.values())


def is_cooked_instance(obj: object) -> bool:
    return hasattr(type(obj), _FIELDS)


def is_cooked_class(cls: object) -> bool:
    if not isinstance(cls, type):
        return False

    return hasattr(cls, _FIELDS)


def is_cooked(obj: object) -> bool:
    cls = obj if isinstance(obj, type) else type(obj)
    return hasattr(cls, _FIELDS)


def _print_err_and_exit(msg: str, out: IO = sys.stderr, exit_code=1, cr: str = "\n"):
    print(f"{msg}{cr}", file=out)
    sys.exit(exit_code)


def _str_to_lines(s: str, width: int) -> List[str]:
    # splits the string into a list
    # of lines, separated at whitespaces
    # so that each line's length is as close
    # to the width as possible
    split_idx = []
    last_space_idx = 0
    for i, c in enumerate(s):
        if c == " ":
            last_space_idx = i
        if i and (i + 1) % width == 0:
            split_idx.append(last_space_idx)

    if not split_idx:
        return [s]

    lines = []
    start_from = 0
    for i in split_idx:
        lines.append(s[start_from:i])
        start_from = i + 1
    lines.append(s[split_idx[-1] + 1 :])
    return lines


def _help_str(
    name: str,
    type_str: str,
    width: int,
    help_padding: int,
    help_str: str,
) -> str:

    name_and_type = f"{name}: {type_str}"
    help_str_width = width - len(name_and_type) - help_padding

    help_lines = _str_to_lines(help_str, help_str_width)
    padding = " " * help_padding
    help_lines[0] = f"{name_and_type}{padding}{help_lines[0]}"
    padding = " " * (len(name_and_type) + help_padding)
    for i in range(1, len(help_lines)):
        help_lines[i] = f"{padding}{help_lines[i]}"

    return "\n".join(help_lines)


def _print_help(
    file: IO,
    prog: str,
    description: Optional[str],
    required: Iterable[Ingredient],
    optional: Iterable[Ingredient],
    width: Optional[int],
    max_width: Optional[int],
    prefixes: Optional[List[str]],
):

    if not width:
        cols = get_terminal_size().columns
        width = cols - 2 if cols <= 90 else 60
        if max_width:
            width = min(max_width, width)

    space_rgx = re.compile(r"\s+")
    max_padding = width // 2

    req_help_strs = []
    for ingredient in required:
        type_str = _TYPE_STR[ingredient.type]
        help_msg = "" if not ingredient.help else space_rgx.sub(" ", ingredient.help)

        if not help_msg:
            req_help_strs.append(f"{ingredient.name}: {type_str}")
        else:
            req_help_strs.append(
                _help_str(
                    ingredient.name,
                    type_str,
                    width,
                    max_padding - len(ingredient.name) - len(type_str),
                    help_msg,
                )
            )

    opt_help_strs = []
    if prefixes:
        for ingredient in optional:
            name = ", ".join(f"{p}{ingredient.name}" for p in prefixes)
            if ingredient.alias:
                alias = ", ".join(f"{p}{ingredient.alias}" for p in prefixes)
                name += f", {alias}"

            type_str = _TYPE_STR[ingredient.type]
            help_msg = (
                "" if not ingredient.help else space_rgx.sub(" ", ingredient.help)
            )

            if not help_msg:
                opt_help_strs.append(f"{name}: {type_str}")
            else:
                opt_help_strs.append(
                    _help_str(
                        name,
                        type_str,
                        width,
                        max_padding - len(name) - len(type_str),
                        help_msg,
                    )
                )
    else:
        for ingredient in optional:
            type_str = _TYPE_STR[ingredient.type]
            help_msg = (
                "" if not ingredient.help else space_rgx.sub(" ", ingredient.help)
            )

            if not help_msg:
                opt_help_strs.append(f"{ingredient.name}: {type_str}")
            else:
                opt_help_strs.append(
                    _help_str(
                        ingredient.name,
                        type_str,
                        width,
                        max_padding - len(ingredient.name) - len(type_str),
                        help_msg,
                    )
                )

    sep = "-" * width
    if description:
        prog = f"{prog}: {description}"

    prog = "\n".join(_str_to_lines(prog, width))

    file.write(f"\n{prog}\n{sep}\n")

    if req_help_strs:
        file.write(f"Required\n{sep}\n")
        file.write("\n".join(req_help_strs) + "\n")

    if opt_help_strs:
        file.write(f"Optional\n{sep}\n")
        file.write("\n".join(opt_help_strs) + "\n")


def _match_opt_strings(
    names: List[str], prefixes: List[str], args_in: List[str]
) -> Iterator[Tuple[int, str]]:
    # iterates through a list of arguments
    # and yields the index and the option string (excluding the prefix)
    # whenever there's a match

    name_group = r"|".join(names) if len(names) > 1 else names[0]
    prefix_group = (
        r"|".join(_ALLOWED_PREFIXES[p] for p in prefixes)
        if len(prefixes) > 1
        else _ALLOWED_PREFIXES[prefixes[0]]
    )
    opt_rgx = re.compile(rf"^({prefix_group})({name_group})$")

    for i, arg in enumerate(args_in):
        match = opt_rgx.match(arg)
        if match is not None:
            yield (i, match.groups()[1])


def serve(
    recipe: Type[CookedBirria],
    raw_ingredients: List[str] = None,
    prefixes: List[str] = None,
    extra_prefixes: List[str] = None,
    description: str = None,
    width: int = None,
    max_width: int = None,
) -> CookedBirria:
    """Parses argument from the command line.

    Takes a class decorated with the cook decorator as
    a list of declarations of arguments. Returns the parsed
    values from command line or argument list in an instance
    of the "cooked" class.

    Args:
        recipe: Class decorated with cook().
        raw_ingredients: Optional list of arguments. Defaults to sys.argv.
        prefixes: List of prefix characters for option strings. Defaults to ['-'].
        extra_prefixes: Extra prefixes characters. Useful for adding extra
            characters to the default list.
        description: Optional description of the program. This is printed out
            along with the help strings for individual arguments.
        width: Optional integer. This is the width of the help message. If not
            provided, the width will be set to a value based on the current terminal size.
        max_width: Optional integer. This is the maximum width (in columns)
            of the program's help message, useful if you don't want to specify a width
            but want to set a boundary for the message. This has no effect
            if width is set.

    Returns:
        Instance of a "cooked" class with parsed values.
    """
    if not is_cooked(recipe):
        raise ValueError(
            "Argument arg_cls must be either a PyArg class or an instance of an PyArg class"
        )

    # use slice to avoid modifying the original list
    if raw_ingredients is None:
        # exclude the first argument (the program invocation)
        preprepped_ingredients = sys.argv[1:]
        prog = PurePath(sys.argv[0]).name
    else:
        preprepped_ingredients = raw_ingredients[:]
        prog = "prog"

    opt_ingredients: Dict[str, Ingredient] = {}
    # map aliasses to real ingredient names
    # when applicable
    opt_ingredients_aliases: Dict[str, str] = {}
    req_ingredients: List[Ingredient] = []
    cooking: Dict[str, CookedIngredient] = {}

    # check for type annotation as well as real values (best effort)
    # if class attr exists or field type is boolean (flag),
    # it's a named argument

    # boolean is always a named argument (aka a flag)
    # if the default for a boolean is set to a boolean,
    # the behaviour is to reverse that value if the
    # corresponding flag is present in the argument list.
    # In all other cases (no default set, default not set to
    # a boolean), set the value to True if the flag is present.

    # list types are a bit more tricky. First of all
    # multiple list fields are basically impossible to
    # handle if they are not optional (and therefore preceded
    # by an option string). So if there are list arguments,
    # the restrictions are:
    # 1. If a non-default field specifies a list type, only one positional
    # argument is allowed (all items parsed not preceded by an option string
    # will be gathered into a list)
    # in addition to other default fields
    # 2. Following 1), the ordering of argument is now "strict", meaning
    # that positional arguments come first in the argument list, followed
    # by named arguments
    # 3. If a default field specifies a list type, and provides a default
    # list through default_factory, any corresponding parsed items will
    # be added to the default list.
    # 4. Following 3), if the field specifies a default value that's not
    # a list, the default value will be overridden by a list of the parsed
    # items.

    seen_req_list = False
    for f in ingredients(recipe):
        if f.name in _HELP_OPT:
            raise TypeError(f"{f.name} is reserved as help option")

        default = getattr(recipe, f.name, MISSING)
        if default is MISSING:
            if f.type == bool:
                # if no default, set the "parsed" value
                # to False here so we can reverse it later
                opt_ingredients[f.name] = f
                cooking[f.name] = False
            elif f.type in _SUPPORTED_LIST_TYPES:
                # list type and default_factory exists -> default field
                if f.default_factory is not MISSING:
                    opt_ingredients[f.name] = f
                else:
                    if seen_req_list:
                        raise TypeError(
                            f"{f.name} cannot be a list,"
                            " only one non-default ingredient can be list"
                        )
                    seen_req_list = True
                    req_ingredients.append(f)
            else:
                if seen_req_list:
                    raise TypeError(
                        "Only one non-default field allowed if using list type,"
                        " use a default field if you want to specify more list fields"
                    )
                req_ingredients.append(f)
        else:
            if f.type == bool:
                # if default is a bool, add it to the parsed
                # dictionary, else set it to False
                if type(default) == bool:
                    cooking[f.name] = default
                else:
                    cooking[f.name] = False
            opt_ingredients[f.name] = f

    num_req = len(req_ingredients)
    if not prefixes:
        prefixes = _DEFAULT_PREFIXES

    if extra_prefixes:
        prefixes += extra_prefixes

    for p in prefixes:
        if p not in _ALLOWED_PREFIXES:
            raise ValueError(f"'{p}' not supported as a prefix")

    prefix_group = r"|".join(_ALLOWED_PREFIXES[p] for p in prefixes)
    help_opt_group = r"|".join(_HELP_OPT)
    help_opt_rgx = re.compile(rf"^({prefix_group})({help_opt_group})$")

    num_raw_ingredients = len(preprepped_ingredients)

    # print help if the help option string is the only argument
    if num_raw_ingredients == 1 and help_opt_rgx.match(preprepped_ingredients[0]):
        _print_help(
            sys.stderr,
            prog,
            description,
            req_ingredients,
            opt_ingredients.values(),
            width,
            max_width,
            prefixes,
        )
        sys.exit(0)

    # just some sanity checks
    if num_req and not num_raw_ingredients:
        _print_err_and_exit(f"No arguments! Need at least {num_req}")

    if not seen_req_list and num_raw_ingredients < num_req:
        _print_err_and_exit(f"Too few arguments! Need at least {num_req}")

    if num_req > 1 and seen_req_list:
        raise TypeError(
            "Only one non-default field allowed if using list type,"
            "Use a default field if you want to specify more list fields"
        )

    if not num_raw_ingredients:
        return recipe(**cooking)  # type: ignore[call-arg]

    # if the first argument is a named argument,
    # all the postional arguments will follow the last
    # named argument. Likewise, if the first argument
    # is a positional argument, all the named arguments
    # follow the last positional argument.
    # Force this convention to make our life easier
    name_list = []
    for n in opt_ingredients.keys():
        name_list.append(n)
        if "_" in n:
            dash_name = "-".join(n.split("_"))
            name_list.append(dash_name)
            opt_ingredients_aliases[dash_name] = n
        alias = opt_ingredients[n].alias
        if alias:
            if alias == n:
                raise ValueError(f"Alias {alias} is the same as ingredient name")
            opt_ingredients_aliases[alias] = n
            name_list.append(alias)

    name_group = "|".join(name_list)
    opt_rgx = re.compile(rf"^({prefix_group})({name_group})$")
    parsed_opts = set()
    first_opt_idx = -1
    last_opt_idx = num_raw_ingredients

    opt_first = bool(opt_rgx.match(preprepped_ingredients[0]))
    if opt_first:
        # non-default field is a list, all named arguments
        # must be at the end of the list
        if seen_req_list:
            _print_err_and_exit(
                "When using list type for non-default field"
                "All default arguments must be specified after the non-default argument"
            )

        first_opt_idx = 0
        last_opt_idx -= num_req

    if opt_ingredients:
        cur = 0
        while cur < num_raw_ingredients:
            arg = preprepped_ingredients[cur]
            match = opt_rgx.match(arg)
            if not match:
                cur += 1
                continue

            name = match.groups()[1]
            if first_opt_idx == -1:
                first_opt_idx = cur
            if name not in opt_ingredients:
                name = opt_ingredients_aliases[name]
            if name in parsed_opts:
                _print_err_and_exit(f"Duplicate instance of {name}")
            parsed_opts.add(name)

            ingredient = opt_ingredients[name]
            itype = ingredient.type
            if itype not in _SUPPORTED_LIST_TYPES:
                if itype == bool:
                    cooking[name] = not cooking[name]
                    cur += 1
                    continue
                else:
                    try:
                        cooking[name] = itype(preprepped_ingredients[cur + 1])
                    except ValueError:
                        _print_err_and_exit(f"Wrong type for {name}, needs {itype}")
                    cur += 2
                    continue
            else:
                vals = []
                cast = _LIST_CASTS[itype]
                for item in preprepped_ingredients[cur + 1 : last_opt_idx]:
                    if opt_rgx.match(item):
                        break
                    try:
                        vals.append(cast(item))
                    except ValueError:
                        _print_err_and_exit(
                            f"value {item} for {name} has the wrong type, needs {cast}"
                        )
                if vals:
                    if ingredient.default_factory is not MISSING:
                        def_list = ingredient.default_factory()
                        if isinstance(def_list, list):
                            cooking[name] = def_list + vals
                        else:
                            cooking[name] = vals
                    else:
                        cooking[name] = vals
                else:
                    _print_err_and_exit(f"Needs at least one value for {name}")
                cur += len(vals) + 1

    if not opt_ingredients or not parsed_opts:
        # no named arguments (either from the class definition or from the list),
        # just parse the list for positional argument
        if not seen_req_list:
            for i, arg in enumerate(preprepped_ingredients):
                try:
                    ingredient = req_ingredients[i]
                    cooking[ingredient.name] = ingredient.type(arg)
                except IndexError:
                    _print_err_and_exit(f"Too many arguments, only need {num_req}")
                except ValueError:
                    _print_err_and_exit(
                        f"Value {arg} of the wrong type, needs {ingredient.type}"
                    )
            return recipe(**cooking)  # type: ignore[call-arg]
        else:
            vals = []
            ingredient = req_ingredients[0]
            cast = _LIST_CASTS[ingredient.type]
            try:
                for arg in preprepped_ingredients:
                    vals.append(cast(arg))
                if not vals:
                    _print_err_and_exit(
                        f"Need at least one value for {ingredient.name}"
                    )
            except ValueError:
                _print_err_and_exit(
                    f"Value {arg} for {ingredient.name} of the wrong type, needs {cast}"
                )

            return recipe(vals)  # type: ignore[call-arg]

    # if the first named arg appears first in the argument list,
    # all the positional arguments are at the end of the list,
    # otherwise they are at the start of the list.

    if num_req:
        pos_args_slice: List[str]
        if not opt_first:
            # if there are named arguments and the first one
            # doesn't start at at the beginning, start from the
            # beginning.
            pos_args_slice = preprepped_ingredients[:num_req]
            # named arguments start from the last positional argument
        else:
            # named arguments start at the beginning, go to
            # back of the list
            pos_args_slice = preprepped_ingredients[-num_req:]
            # named arguments go until the first positional argument

        if not seen_req_list:
            for i, arg in enumerate(pos_args_slice):
                try:
                    ingredient = req_ingredients[i]
                    cooking[ingredient.name] = ingredient.type(arg)
                except IndexError:
                    _print_err_and_exit(f"Need positinal argument: {ingredient.name}")
                except ValueError:
                    _print_err_and_exit(
                        f"Value {arg} for {ingredient.name} of the wrong type, needs {ingredient.type}"
                    )
        else:
            vals = []
            ingredient = req_ingredients[0]
            cast = _LIST_CASTS[ingredient.type]
            try:
                for arg in preprepped_ingredients[:first_opt_idx]:
                    # gather items into a list until we find an option string
                    vals.append(cast(arg))
                if not vals:
                    _print_err_and_exit(
                        f"Need at least one value for {ingredient.name}"
                    )
            except ValueError:
                _print_err_and_exit(
                    f"Value {arg} for {ingredient.name} of the wrong type, needs {cast}"
                )
            cooking[ingredient.name] = vals

    return recipe(**cooking)  # type: ignore[call-arg]



import itertools
from typing import List, Tuple, TypeVar, Type, Callable, Any, Union

import pytest
from pytest import CaptureFixture

from birria import (
    cook,
    ingredient,
    ingredients,
    is_cooked,
    is_cooked_instance,
    is_cooked_class,
    serve,
    _match_opt_strings,
)


S = TypeVar("S", int, str, float)
AllowedList = List[S]
IngredientType = Union[
    int, float, str, bool, list, List[str], List[int], List[float], List
]

CAST_LOOKUP = {
    list: str,
    List: str,
    List[str]: str,
    List[float]: float,
    List[int]: int,
}


def arg_strs_to_list(a1: str, *args: str) -> List[str]:
    raw_list = [a1] if a1 else []
    raw_list += [a for a in args if a]
    all_args = " ".join(raw_list)
    return all_args.split(" ")


def assert_exit(
    exit_code: int, exit_func: Callable[..., Any], *args: Any, **kwargs: Any
):
    with pytest.raises(SystemExit) as wrapped_err:
        exit_func(*args, **kwargs)
    assert wrapped_err.type == SystemExit
    assert wrapped_err.value.code == exit_code


def test_cooking_functions():
    @cook
    class CookedClass:
        a: int
        b: int
        c: str
        d: List[float]
        e: bool = False

    assert is_cooked_class(CookedClass)
    assert is_cooked(CookedClass)
    assert not is_cooked_instance(CookedClass)

    cooked = CookedClass(1, 2, "some string", [0.4, 1.2])
    assert is_cooked_instance(cooked)
    assert is_cooked(cooked)
    assert not is_cooked_class(cooked)

    class_ingredients = ingredients(CookedClass)
    instance_ingredients = ingredients(cooked)
    assert class_ingredients == instance_ingredients


def test_cook_class_empty():
    @cook
    class BaseRecipe:
        a: int
        b: float
        d: bool

    with pytest.raises(TypeError):

        @cook
        class EmptyRecipe:
            pass

    @cook
    class DerivedRecipe(BaseRecipe):
        pass

    assert ingredients(DerivedRecipe) == ingredients(BaseRecipe)


@pytest.mark.parametrize(
    "int_val, float_val, str_val, bool_val, int_list, float_list, str_list",
    [
        (
            1,
            2.0,
            "hello",
            False,
            [1, 2, 3],
            [1.0, 2.0, 3.0, 4.5],
            ["some", "lame", "string"],
        ),
        (
            -199,
            -9.1231,
            "",
            True,
            [i for i in range(100)],
            [-0.5 * i for i in range(20, 50, 3)],
            ["a b c d e f".split()],
        ),
        (
            2 ** 64,
            3.142,
            ",".join((str(i) for i in range(100))),
            True,
            [1, 2, 3],
            None,
            [],
        ),
    ],
)
def test_cook_class_init_no_default(
    int_val: int,
    float_val: float,
    str_val: str,
    bool_val: bool,
    int_list: List[int],
    float_list: List[float],
    str_list: List[str],
):
    @cook
    class CookedClass:
        a: int
        b: float
        c: str
        d: bool
        e: List[int]
        f: List[float]
        g: List[str]

    cooked = CookedClass(
        int_val,
        float_val,
        str_val,
        bool_val,
        int_list,
        float_list,
        str_list,
    )
    assert cooked.a == int_val
    assert cooked.b == float_val
    assert cooked.c == str_val
    assert cooked.d == bool_val
    assert cooked.e == int_list
    assert cooked.f == float_list
    assert cooked.g == str_list


@pytest.mark.parametrize(
    "int_val, float_val, str_val, bool_val, int_list, float_list, str_list",
    [
        (
            1,
            2.0,
            "hello",
            False,
            [1, 2, 3],
            [1.0, 2.0, 3.0, 4.5],
            ["some", "lame", "string"],
        ),
        (
            -199,
            -9.1231,
            "",
            True,
            [i for i in range(100)],
            [-0.5 * i for i in range(20, 50, 3)],
            ["a b c d e f".split()],
        ),
        (
            2 ** 64,
            3.142,
            ",".join((str(i) for i in range(100))),
            True,
            [1, 2, 3],
            None,
            [],
        ),
    ],
)
def test_cook_class_all_default(
    int_val: int,
    float_val: float,
    str_val: str,
    bool_val: bool,
    int_list: List[int],
    float_list: List[float],
    str_list: List[str],
):
    @cook
    class CookedClass:
        a: int = int_val
        b: float = float_val
        c: str = str_val
        d: bool = bool_val
        e: List[int] = ingredient(default_factory=lambda: int_list)
        f: List[float] = ingredient(default_factory=lambda: float_list)
        g: List[str] = ingredient(default_factory=lambda: str_list)

    cooked = CookedClass()
    assert cooked.a == int_val
    assert cooked.b == float_val
    assert cooked.c == str_val
    assert cooked.d == bool_val
    assert cooked.e == int_list
    assert cooked.f == float_list
    assert cooked.g == str_list


@pytest.mark.parametrize(
    "int_val, float_val, str_val, bool_val, int_list, float_list, str_list",
    [
        (
            1,
            2.0,
            "hello",
            False,
            [1, 2, 3],
            [1.0, 2.0, 3.0, 4.5],
            ["some", "lame", "string"],
        ),
        (
            -199,
            -9.1231,
            "",
            True,
            [i for i in range(100)],
            [-0.5 * i for i in range(20, 50, 3)],
            ["a b c d e f".split()],
        ),
        (
            2 ** 64,
            3.142,
            ",".join((str(i) for i in range(100))),
            True,
            [1, 2, 3],
            None,
            [],
        ),
    ],
)
def test_cook_class_init_mixed_ordered(
    int_val: int,
    float_val: float,
    str_val: str,
    bool_val: bool,
    int_list: List[int],
    float_list: List[float],
    str_list: List[str],
):
    @cook
    class CookedClass:
        a: int
        b: float
        c: str
        d: bool = bool_val
        e: List[int] = ingredient(default_factory=lambda: int_list)
        f: List[float] = ingredient(default_factory=lambda: float_list)
        g: List[str] = ingredient(default_factory=lambda: str_list)

    cooked = CookedClass(int_val, float_val, str_val)
    assert cooked.a == int_val
    assert cooked.b == float_val
    assert cooked.c == str_val
    assert cooked.d == bool_val
    assert cooked.e == int_list
    assert cooked.f == float_list
    assert cooked.g == str_list


@pytest.mark.parametrize(
    "int_val, float_val, str_val, bool_val, int_list, float_list, str_list",
    [
        (
            1,
            2.0,
            "hello",
            False,
            [1, 2, 3],
            [1.0, 2.0, 3.0, 4.5],
            ["some", "lame", "string"],
        ),
        (
            -199,
            -9.1231,
            "",
            True,
            [i for i in range(100)],
            [-0.5 * i for i in range(20, 50, 3)],
            ["a b c d e f".split()],
        ),
        (
            2 ** 64,
            3.142,
            ",".join((str(i) for i in range(100))),
            True,
            [1, 2, 3],
            None,
            [],
        ),
    ],
)
def test_cook_class_init_mixed_unordered(
    int_val: int,
    float_val: float,
    str_val: str,
    bool_val: bool,
    int_list: List[int],
    float_list: List[float],
    str_list: List[str],
):
    @cook
    class CookedClass:
        a: int
        b: float = float_val
        c: str
        d: bool
        e: List[int] = ingredient(default_factory=lambda: int_list)
        f: List[float]
        g: List[str]

    # test for proper ordering of positional argument
    # regardless of the order of the ingredient/field
    # declaration, the generated __init__ should
    # have proper argument ordering: positional (non-default)
    # followed by named arguments (default)
    # in this case, the proper order is
    # a, c, d, f, g (positional) then
    # b, e, f (named)

    # test for __init__ with just the postional argument
    cooked = CookedClass(int_val, str_val, bool_val, float_list, str_list)
    assert cooked.a == int_val
    assert cooked.b == float_val
    assert cooked.c == str_val
    assert cooked.d == bool_val
    assert cooked.e == int_list
    assert cooked.f == float_list
    assert cooked.g == str_list


@pytest.mark.parametrize(
    "int_val_tup, float_val_tup, str_val_tup, bool_val_tup, int_list_tup, float_list_tup, str_list_tup",
    [
        (
            (1, -1),
            (2.0, 0.0),
            ("hello", "goodbye"),
            (True, False),
            ([1, 2, 3], [3, 2, 1]),
            ([1.0, 2.0, 3.0, 4.5], [0.0]),
            (["some", "lame", "string"], ["some", "cool", "string"]),
        ),
        (
            (-199, None),
            (-9.1231, None),
            ("", None),
            (True, None),
            ([i for i in range(100)], None),
            ([-0.5 * i for i in range(20, 50, 3)], None),
            (["a b c d e f".split()], None),
        ),
        (
            (2 ** 64, 2 ** 64),
            (3.142, 3.142),
            (
                ",".join((str(i) for i in range(100))),
                ",".join((str(i) for i in range(100))),
            ),
            (True, True),
            ([1, 2, 3], [1, 2, 3]),
            (None, None),
            ([], []),
        ),
    ],
)
def test_cook_class_init_default_override(
    int_val_tup: Tuple[int, int],
    float_val_tup: Tuple[float, float],
    str_val_tup: Tuple[str, str],
    bool_val_tup: Tuple[bool, bool],
    int_list_tup: Tuple[List[int], List[int]],
    float_list_tup: Tuple[List[float], List[float]],
    str_list_tup: Tuple[List[str], List[str]],
):

    def_int, override_int = int_val_tup
    def_float, override_float = float_val_tup
    def_str, override_str = str_val_tup
    def_bool, override_bool = bool_val_tup
    def_int_list, override_int_list = int_list_tup
    def_float_list, override_float_list = float_list_tup
    def_str_list, override_str_list = str_list_tup

    @cook
    class CookedClass:
        a: int = def_int
        b: float = def_float
        c: str = def_str
        d: bool = def_bool
        e: List[int] = ingredient(default_factory=lambda: def_int_list)
        f: List[float] = ingredient(default_factory=lambda: def_float_list)
        g: List[str] = ingredient(default_factory=lambda: def_str_list)

    cooked_boring = CookedClass()
    assert cooked_boring.a == def_int
    assert cooked_boring.b == def_float
    assert cooked_boring.c == def_str
    assert cooked_boring.d == def_bool
    assert cooked_boring.e == def_int_list
    assert cooked_boring.f == def_float_list
    assert cooked_boring.g == def_str_list

    cooked_fun = CookedClass(
        a=override_int,
        b=override_float,
        c=override_str,
        d=override_bool,
        e=override_int_list,
        f=override_float_list,
        g=override_str_list,
    )
    assert cooked_fun.a == override_int
    assert cooked_fun.b == override_float
    assert cooked_fun.c == override_str
    assert cooked_fun.d == override_bool
    assert cooked_fun.e == override_int_list
    assert cooked_fun.f == override_float_list
    assert cooked_fun.g == override_str_list


def test_cook_inheritence_no_override():
    @cook
    class Grandma:
        gran_int: int
        gran_float: float
        gran_str: str

    @cook
    class Ma(Grandma):
        ma_list_int: List[int]
        ma_bool: bool

    @cook
    class Me(Ma):
        me_list_float: List[float]

    grandma = Grandma(-1, 0.0, "grandma")
    assert grandma.gran_int == -1
    assert grandma.gran_float == 0.0
    assert grandma.gran_str == "grandma"

    ma = Ma(1, 0.5, "ma", [1, 2, 3], True)
    assert ma.gran_int == 1
    assert ma.gran_float == 0.5
    assert ma.gran_str == "ma"
    assert ma.ma_list_int == [1, 2, 3]
    assert ma.ma_bool

    me = Me(2, 3.142, "me", [3, 2, 1], False, [0.5, 0.25, 0.125])
    assert me.gran_int == 2
    assert me.gran_float == 3.142
    assert me.gran_str == "me"
    assert me.ma_list_int == [3, 2, 1]
    assert not me.ma_bool
    assert me.me_list_float == [0.5, 0.25, 0.125]

    # test for ingredients of derived class
    # includes the ingredients declared
    # in base classes

    grandma_ingredients = ingredients(grandma)
    ma_ingredients = ingredients(ma)
    me_ingredients = ingredients(me)
    for i in grandma_ingredients:
        assert i in ma_ingredients
    for i in ma_ingredients:
        assert i in me_ingredients


def test_cook_inheritence_override():
    @cook
    class Grandma:
        gran_int: int
        gran_float: float
        gran_str: str = "grandma"

    @cook
    class Ma(Grandma):
        ma_list_int: List[int]
        ma_bool: bool = True
        gran_str: str = "ma"

    @cook
    class Me(Ma):
        ma_bool: bool = False
        me_list_float: List[float]
        gran_str: str = "me"

    # test override parent classes' attributes
    assert Grandma.gran_str == "grandma"
    assert Ma.ma_bool
    assert Ma.gran_str == "ma"
    assert Me.gran_str == "me"
    assert not Me.ma_bool

    # test override on instance's attributes
    grandma = Grandma(-1, 0.0)
    assert grandma.gran_int == -1
    assert grandma.gran_float == 0.0
    assert grandma.gran_str == "grandma"

    ma = Ma(1, 0.5, [1, 2, 3])
    assert ma.gran_int == 1
    assert ma.gran_float == 0.5
    assert ma.gran_str == "ma"
    assert ma.ma_list_int == [1, 2, 3]
    assert ma.ma_bool

    me = Me(2, 3.142, [3, 2, 1], [0.5, 0.25, 0.125])
    assert me.gran_int == 2
    assert me.gran_float == 3.142
    assert me.gran_str == "me"
    assert me.ma_list_int == [3, 2, 1]
    assert not me.ma_bool
    assert me.me_list_float == [0.5, 0.25, 0.125]

    # test if the overrides show up in ingredients()
    grandma_ingredients = {i.name: i for i in ingredients(grandma)}
    ma_ingredients = {i.name: i for i in ingredients(ma)}
    me_ingredients = {i.name: i for i in ingredients(me)}

    assert grandma_ingredients["gran_str"].default != ma_ingredients["gran_str"].default
    assert grandma_ingredients["gran_str"].default != me_ingredients["gran_str"].default
    assert ma_ingredients["gran_str"].default != me_ingredients["gran_str"].default
    assert ma_ingredients["ma_bool"].default != me_ingredients["ma_bool"].default

    for i in grandma_ingredients:
        assert i in ma_ingredients
        assert i in me_ingredients


@pytest.mark.parametrize(
    "raw_ingredients",
    [
        ["1", "2.0", "hello"],
        ["0", "-1231.123", "some lame string"],
        ["0", "-1231.123", "some lame string"],
    ],
)
def test_serve_positional(
    raw_ingredients: List[str],
):
    @cook
    class CookedClass:
        a: int
        b: float
        c: str

    served_birria = serve(CookedClass, raw_ingredients)
    exp_int, exp_float, exp_str = raw_ingredients

    assert served_birria.a == int(exp_int)
    assert served_birria.b == float(exp_float)
    assert served_birria.c == exp_str

    # does parsing still work with extra items
    # probably should just fail for now
    assert_exit(1, serve, CookedClass, raw_ingredients + [1, 2, 3])

    # cases where we fail with exit code 1
    assert_exit(1, serve, CookedClass, [])
    num_args = len(raw_ingredients)
    for i in range(num_args - 1):
        # trim the list one by one forward and backward
        assert_exit(1, serve, CookedClass, raw_ingredients[i + 1 :])
        assert_exit(1, serve, CookedClass, raw_ingredients[: -i - 1])


@pytest.mark.parametrize(
    "scalar_int, scalar_float, scalar_str, bool_flag, list_int, list_float, list_str",
    [
        (
            "-a 1",
            "-b 3.142",
            "-c some",
            "-d",
            "-e 1 2 3",
            "-f 1.0 2.0 3.0",
            "-g some lame string",
        ),
        (
            "-a -100",
            "-b 0.5",
            "-c None",
            "",
            "-e -1 -2 -3 -4 -5",
            "-f 0.0",
            "-g some nice string",
        ),
        (
            "-a -100",
            "-b 0.5",
            "-c -c",
            "",
            "-e -1 -2 -3 -4 -5",
            "-f 0.0",
            "-g some nice string",
        ),
    ],
)
def test_serve_named(
    scalar_int: str,
    scalar_float: str,
    scalar_str: str,
    bool_flag: str,
    list_int: str,
    list_float: str,
    list_str: str,
):
    @cook
    class CookedClass:
        a: int = None
        b: float = None
        c: str = None
        d: bool = None
        e: List[int] = None
        f: List[float] = None
        g: List[str] = None

    cooked_w_no_seasoning = CookedClass()
    assert cooked_w_no_seasoning.a is None
    assert cooked_w_no_seasoning.b is None
    assert cooked_w_no_seasoning.c is None
    assert cooked_w_no_seasoning.d is None
    assert cooked_w_no_seasoning.e is None
    assert cooked_w_no_seasoning.f is None
    assert cooked_w_no_seasoning.g is None

    served_w_no_seasoning = serve(CookedClass, raw_ingredients=[])
    assert served_w_no_seasoning.a is None
    assert served_w_no_seasoning.b is None
    assert served_w_no_seasoning.c is None
    # default behaviour for boolean flags
    # if no default boolean value is provided
    # is to set it to False
    assert served_w_no_seasoning.d is False
    assert served_w_no_seasoning.e is None
    assert served_w_no_seasoning.f is None
    assert served_w_no_seasoning.g is None

    all_args = [
        scalar_int,
        scalar_float,
        scalar_str,
        bool_flag,
        list_int,
        list_float,
        list_str,
    ]

    # test with regular order
    arg_list = arg_strs_to_list(*all_args)

    exp_int = int(scalar_int.split(" ")[1])
    exp_float = float(scalar_float.split(" ")[1])
    exp_str = scalar_str.split(" ")[1]
    exp_bool = bool_flag == "-d"
    exp_int_list = [int(i) for i in list_int.split(" ")[1:]]
    exp_float_list = [float(i) for i in list_float.split(" ")[1:]]
    exp_str_list = [i for i in list_str.split(" ")[1:]]

    served_seasoned = serve(CookedClass, raw_ingredients=arg_list)
    assert served_seasoned.a == exp_int
    assert served_seasoned.b == exp_float
    assert served_seasoned.c == exp_str
    assert served_seasoned.d == exp_bool
    assert served_seasoned.e == exp_int_list
    assert served_seasoned.f == exp_float_list
    assert served_seasoned.g == exp_str_list

    # test with permutations
    for arg_perm in itertools.permutations(all_args):
        arg_list = arg_strs_to_list(*arg_perm)
        served_seasoned_scrambled = serve(CookedClass, raw_ingredients=arg_list)
        assert served_seasoned_scrambled.a == exp_int
        assert served_seasoned_scrambled.b == exp_float
        assert served_seasoned_scrambled.c == exp_str
        assert served_seasoned_scrambled.d == exp_bool
        assert served_seasoned_scrambled.e == exp_int_list
        assert served_seasoned_scrambled.f == exp_float_list
        assert served_seasoned_scrambled.g == exp_str_list


@pytest.mark.parametrize(
    "scalar_int, scalar_float, scalar_str, bool_flag, list_int, list_float, list_str",
    [
        (
            "1",
            "3.142",
            "some",
            "-d",
            "-e 1 2 3",
            "-f 1.0 2.0 3.0",
            "-g some lame string",
        ),
        (
            "-100",
            "0.5",
            "None",
            "",
            "-e -1 -2 -3 -4 -5",
            "-f 0.0",
            "-g some nice string",
        ),
    ],
)
def test_serve_mixed(
    scalar_int: str,
    scalar_float: str,
    scalar_str: str,
    bool_flag: str,
    list_int: str,
    list_float: str,
    list_str: str,
):
    @cook
    class CookedClass:
        a: int
        b: float
        c: str
        d: bool = True
        e: List[int] = None
        f: List[float] = None
        g: List[str] = None

    exp_int = int(scalar_int)
    exp_float = float(scalar_float)
    exp_str = scalar_str
    exp_bool = not (bool_flag == "-d")
    exp_int_list = [int(i) for i in list_int.split(" ")[1:]]
    exp_float_list = [float(i) for i in list_float.split(" ")[1:]]
    exp_str_list = [i for i in list_str.split(" ")[1:]]

    all_pos_args = [scalar_int, scalar_float, scalar_str]
    all_named_args = [bool_flag, list_int, list_float, list_str]

    cooked_w_no_seasoning = CookedClass(exp_int, exp_float, exp_str)
    assert cooked_w_no_seasoning.a == exp_int
    assert cooked_w_no_seasoning.b == exp_float
    assert cooked_w_no_seasoning.c == exp_str
    assert cooked_w_no_seasoning.d
    assert cooked_w_no_seasoning.e is None
    assert cooked_w_no_seasoning.f is None
    assert cooked_w_no_seasoning.g is None

    pos_arg_list = [scalar_int, scalar_float, scalar_str]
    served_w_no_seasoning = serve(CookedClass, raw_ingredients=all_pos_args)
    assert served_w_no_seasoning.a == exp_int
    assert served_w_no_seasoning.b == exp_float
    assert served_w_no_seasoning.c == exp_str
    assert served_w_no_seasoning.d
    assert served_w_no_seasoning.e is None
    assert served_w_no_seasoning.f is None
    assert served_w_no_seasoning.g is None

    # test all permutations for named arguments
    for perm_narg in itertools.permutations(all_named_args):
        narg_list = arg_strs_to_list(*perm_narg)

        # pos followed by narg
        args = pos_arg_list + narg_list
        served_seasoned_scrambled = serve(CookedClass, raw_ingredients=args)
        assert served_seasoned_scrambled.a == exp_int
        assert served_seasoned_scrambled.b == exp_float
        assert served_seasoned_scrambled.c == exp_str
        assert served_seasoned_scrambled.d == exp_bool
        assert served_seasoned_scrambled.e == exp_int_list
        assert served_seasoned_scrambled.f == exp_float_list
        assert served_seasoned_scrambled.g == exp_str_list

        # narg followed by pos
        args = narg_list + pos_arg_list
        served_seasoned_scrambled = serve(CookedClass, raw_ingredients=args)
        assert served_seasoned_scrambled.a == exp_int
        assert served_seasoned_scrambled.b == exp_float
        assert served_seasoned_scrambled.c == exp_str
        assert served_seasoned_scrambled.d == exp_bool
        assert served_seasoned_scrambled.e == exp_int_list
        assert served_seasoned_scrambled.f == exp_float_list
        assert served_seasoned_scrambled.g == exp_str_list


@pytest.mark.parametrize(
    "default, more, itype",
    [
        ([], ["some", "lame", "string"], List[str]),
        ([1, 2, 3], [str(i) for i in range(10)], List[int]),
        (None, ["0.5", "0.25", "0.125"], List[float]),
        (None, ["asdfas", "fdasf", "fcaed"], List[int]),
        (None, ["asdfas", "fdasf", "fcaed"], List[float]),
    ],
)
def test_serve_list_default(default: Any, more: AllowedList, itype: Type[AllowedList]):
    @cook
    class CookedClass:
        shopping_list: itype = ingredient(default_factory=lambda: default)

    served_w_no_seasoning = serve(CookedClass, raw_ingredients=[])
    assert served_w_no_seasoning.shopping_list == default

    assert_exit(1, serve, CookedClass, ["-shopping_list"])

    args = ["-shopping_list"] + more
    cast_inner = CAST_LOOKUP[itype]
    wrong_type = False
    try:
        casted_more = [cast_inner(i) for i in more]
    except ValueError:
        wrong_type = True

    if wrong_type:
        # if a value is of the wrong type (cast failed)
        # we should exit
        assert_exit(1, serve, CookedClass, args)
    else:
        served_seasoned = serve(CookedClass, raw_ingredients=args)

        if isinstance(default, list):
            # if default is provided and it's a list,
            # the parsed items should be added to default
            assert served_seasoned.shopping_list == default + casted_more
        else:
            # otherwise just return the parsed items packed in a list
            assert served_seasoned.shopping_list == casted_more


@pytest.mark.parametrize(
    "args, itype",
    [
        (["1", "2", "3"], List[int]),
        (["some", "lame", "string"], List[str]),
        (["0.5", "0.25", "0.125"], List[float]),
        (["what", "the", "f", "ever"], list),
        (["one"], List),
        (["asdfas", "fdasf", "fcaed"], List[int]),
        (["asdfas", "fdasf", "fcaed"], List[float]),
    ],
)
def test_serve_list_no_default(args: AllowedList, itype: Type[AllowedList]):
    @cook
    class CookedClass:
        shopping_list: itype

    # if a list field is non-default,
    # and no corresponding values are
    # found, exit with code 1
    assert_exit(1, serve, CookedClass, [])

    cast = CAST_LOOKUP[itype]
    wrong_type = False

    try:
        expected = [cast(i) for i in args]
    except ValueError:
        wrong_type = True

    if not wrong_type:
        served = serve(CookedClass, raw_ingredients=args)
        assert served.shopping_list == expected
    else:
        assert_exit(1, serve, CookedClass, args)


def test_serve_list_mixed():
    @cook
    class CookedClass:
        a: List[str]
        b: List[int] = None
        c: List[float] = ingredient(default_factory=list)
        d: List[float] = ingredient(default_factory=lambda: [1.0, 2.0, 3.0])

    exp_a = ["some", "lame", "string"]

    served_w_no_seasoning = serve(CookedClass, raw_ingredients=exp_a)

    assert served_w_no_seasoning.a == exp_a
    assert served_w_no_seasoning.b is None
    assert served_w_no_seasoning.c == []
    assert served_w_no_seasoning.d == [1.0, 2.0, 3.0]

    assert_exit(1, serve, CookedClass, exp_a + ["-b"])
    assert_exit(1, serve, CookedClass, exp_a + ["-c"])
    assert_exit(1, serve, CookedClass, exp_a + ["-d"])

    b_vals = ["-b", "1", "2", "3"]
    exp_b = [int(i) for i in b_vals[1:]]
    c_vals = ["-c", "0.5", "0.25", "0.125"]
    exp_c = [float(i) for i in c_vals[1:]]
    d_vals = ["-d", "-1.0", "-2.0", "-3.0"]
    exp_d = [1.0, 2.0, 3.0] + [float(i) for i in d_vals[1:]]

    all_named = [b_vals, c_vals, d_vals]

    # test all the permutations
    for perm in itertools.permutations(all_named):
        flatten = [i for i in itertools.chain.from_iterable(perm)]
        args = exp_a + flatten
        served_seasoned = serve(CookedClass, raw_ingredients=args)
        assert served_seasoned.a == exp_a
        assert served_seasoned.b == exp_b
        assert served_seasoned.c == exp_c
        assert served_seasoned.d == exp_d

        rev_args = flatten + exp_a
        assert_exit(1, serve, CookedClass, rev_args)


def test_serve_bool():
    @cook
    class CookedClass:
        b: bool

    cooked_w_no_seasoning = serve(CookedClass, raw_ingredients=[])
    assert cooked_w_no_seasoning.b is False

    cooked_seasoned = serve(CookedClass, raw_ingredients=["-b"])
    assert cooked_seasoned.b is True

    @cook
    class TrueCookedClass:
        b: bool = True

    cooked_w_no_seasoning = serve(TrueCookedClass, raw_ingredients=[])
    assert cooked_w_no_seasoning.b is True
    cooked_seasoned = serve(TrueCookedClass, raw_ingredients=["-b"])
    assert cooked_seasoned.b is False

    @cook
    class FalseCookedClass:
        b: bool = False

    cooked_w_no_seasoning = serve(FalseCookedClass, raw_ingredients=[])
    assert cooked_w_no_seasoning.b is False
    cooked_seasoned = serve(FalseCookedClass, raw_ingredients=["-b"])
    assert cooked_seasoned.b is True

    @cook
    class WeirdCookedClass:
        b: bool = None

    cooked_w_no_seasoning = serve(WeirdCookedClass, raw_ingredients=[])
    assert cooked_w_no_seasoning.b is False
    cooked_seasoned = serve(WeirdCookedClass, raw_ingredients=["-b"])
    assert cooked_seasoned.b is True


@pytest.mark.parametrize(
    "prefixes, extra_prefixes", [(["-", "--"], None), (None, ["/"]), (["-"], ["+"])]
)
def test_serve_prefix(prefixes: List[str], extra_prefixes: List[str]):
    @cook
    class CookedClass:
        a: int = None
        b: float = None
        c: str = None
        d: List[int] = None

    # allowed prefixes are '-', '/, '+'
    # the parser should be to able
    # to match option strings with the provided prefixes
    # in any arbitrary combination.
    # Something like "-a 1 /b 0.5 +c string -d .....)"
    # should just work

    ingredient_names = ["a", "b", "c", "d"]
    prefixes = prefixes or ["-"]
    if extra_prefixes:
        prefixes += extra_prefixes
    illegal_prefix = not all(p in ("-", "+", "/") for p in prefixes)

    exp_a = 1
    exp_b = 0.5
    exp_c = "hello"
    exp_d = [1, 2, 3]
    exp_lookup = {"a": exp_a, "b": exp_b, "c": exp_c, "d": exp_d}

    # iterate through:
    # -a -b -c -d
    # -a -b -c --d
    # -a -b --c -d
    # -a --b -c -d
    # -a --b -c -d
    # -a --b -c --d
    # -a --b --c -d
    # -a --b --c --d
    # repeats for --a (and then for b, c, d)
    for prefix_combination in itertools.product(prefixes, repeat=len(ingredient_names)):
        arg_strs = []
        for i, p in enumerate(prefix_combination):
            name = ingredient_names[i]
            opt_str = f"{p}{name}"
            val = exp_lookup[name]
            if isinstance(val, list):
                arg_str = f"{opt_str} {' '.join([str(v) for v in val])}"
            else:
                arg_str = f"{opt_str} {val}"

            arg_strs.append(arg_str)

        for perm in itertools.permutations(arg_strs):
            args = arg_strs_to_list(*perm)
            if illegal_prefix:
                # should raise ValueError here
                with pytest.raises(ValueError):
                    served = serve(
                        CookedClass,
                        raw_ingredients=args,
                        prefixes=prefixes,
                        extra_prefixes=extra_prefixes,
                    )
            else:
                served = serve(
                    CookedClass,
                    raw_ingredients=args,
                    prefixes=prefixes,
                    extra_prefixes=extra_prefixes,
                )
                assert served.a == exp_a
                assert served.b == exp_b
                assert served.c == exp_c
                assert served.d == exp_d


def test_serve_snake_case_names():
    @cook
    class Recipe:
        long_name: str = "hello"
        even_longer_name: float = 0.0
        ridiculously_long_name: List[str] = None

    # for snake case names (underscore between the words)
    # such as "long_name", we should accept both
    # "{prefix}long-name" and "{prefix}long_name" as
    # valid option strings
    exp_a = "lame"
    exp_b = 0.5
    exp_c = ["some", "lame", "string"]
    exp_lookup = {
        "long name": exp_a,
        "even longer name": exp_b,
        "ridiculously long name": exp_c,
    }

    deliminators = ["-", "_"]
    ingredient_names_no_delim = [
        "long name",
        "even longer name",
        "ridiculously long name",
    ]

    # cases like -long-name ... long_name
    # should be treated like duplicate
    # instances of the same argument ->
    # print out error and exit
    assert_exit(1, serve, Recipe, ["-long-name", "long", "-long_name", "name"])

    for delim_combo in itertools.product(
        deliminators, repeat=len(ingredient_names_no_delim)
    ):
        arg_strs = []
        for i, delim in enumerate(delim_combo):
            name = ingredient_names_no_delim[i]
            opt_str = f"-{delim.join(name.split())}"
            val = exp_lookup[name]
            if isinstance(val, list):
                arg_str = f"{opt_str} {' '.join([str(v) for v in val])}"
            else:
                arg_str = f"{opt_str} {val}"

            arg_strs.append(arg_str)

        for perm in itertools.permutations(arg_strs):
            args = arg_strs_to_list(*perm)
            served = serve(
                Recipe,
                raw_ingredients=args,
            )
            assert served.long_name == exp_a
            assert served.even_longer_name == exp_b
            assert served.ridiculously_long_name == exp_c


@pytest.mark.parametrize("prefixes", [["-", "+", "/"]])
def test_help(prefixes: List[str], capsys: CaptureFixture[str]):

    # test that "help" and "h" are reserved
    @cook
    class Recipe:
        help: str

    with pytest.raises(TypeError):
        _ = serve(Recipe)

    @cook
    class Recipe:
        h: str

    with pytest.raises(TypeError):
        _ = serve(Recipe)

    @cook
    class Recipe:
        help: str = "help"

    with pytest.raises(TypeError):
        _ = serve(Recipe)

    @cook
    class Recipe:
        h: str = None

    with pytest.raises(TypeError):
        _ = serve(Recipe)

    # test that we print out something to stderr and exit 0
    # when the help option string is the only argument on the cli,
    # don't really need to test for the exact help string
    # as that's subject to change
    @cook
    class Recipe:
        a: int
        b: float
        c: str
        d: bool
        f: List[str] = None

    help_opt_strs = ["".join(s) for s in itertools.product(prefixes, ("h", "help"))]

    for s in help_opt_strs:
        assert_exit(0, serve, Recipe, [s], prefixes)
        captured = capsys.readouterr().err
        assert captured

        # help strings should differ when serve called with optional
        # params (description, width, max_witdh), but if both width
        # and max_width is provided, the resulting help string should be
        # the same.

        assert_exit(0, serve, Recipe, [s], prefixes, None, None, 40)
        captured_w_width = capsys.readouterr().err
        assert captured_w_width
        assert captured_w_width != captured
        assert_exit(0, serve, Recipe, [s], prefixes, None, None, 40, 30)
        captured_w_width_and_max_width = capsys.readouterr().err
        assert captured_w_width == captured_w_width_and_max_width

        assert_exit(0, serve, Recipe, [s], prefixes, None, "test description")
        captured_w_description = capsys.readouterr().err
        assert captured_w_description
        assert captured_w_description != captured
        assert captured_w_description != captured_w_width

        assert_exit(0, serve, Recipe, [s], prefixes, None, "test description", 40)
        captured_w_description_and_width = capsys.readouterr().err
        assert captured_w_description_and_width
        assert captured_w_description_and_width != captured
        assert captured_w_description_and_width != captured_w_width
        assert captured_w_description_and_width != captured_w_description
        assert_exit(0, serve, Recipe, [s], prefixes, None, "test description", 40, 30)
        captured_w_description_and_width_and_max_width = capsys.readouterr().err
        assert (
            captured_w_description_and_width
            == captured_w_description_and_width_and_max_width
        )

    # test that parsing behaves normally if an item is the same
    # as the help option string. The parser should just treat
    # that as any other item in the list.

    exp_a = 1
    exp_b = 2.0
    exp_c = "hello"

    for s in help_opt_strs:
        args = [str(exp_a), str(exp_b), s]
        served = serve(Recipe, raw_ingredients=args)
        assert served.a == exp_a
        assert served.b == exp_b
        assert served.c == s
        args = [str(exp_a), str(exp_b), exp_c, "-f", s, "more", "strings", "-d"]
        served = serve(Recipe, raw_ingredients=args)
        assert served.a == exp_a
        assert served.b == exp_b
        assert served.c == exp_c
        assert served.d
        assert served.f == [s, "more", "strings"]


@pytest.mark.parametrize(
    "prefixes, names, expected, raw_names",
    [
        (["-", "+", "/"], ["b"], [(1, "b")], ["blah", "b", "+bloh", "bbb", "+-b"]),
        (
            ["-"],
            ["blah", "b"],
            [(0, "blah"), (3, "b")],
            ["blah", "+b", "-b", "b", "+-b", "bob"],
        ),
    ],
)
def test_match_opt_string_all_permumations(
    prefixes: List[str],
    names: List[str],
    expected: List[Tuple[int, str]],
    raw_names: List[str],
):
    # test for permutations of option strings (prefixes * opt names)
    # prefixes -> list of possible prefix
    # names -> names to match against
    # expected -> list of expected index, name tuples that the function yields
    # raw_names -> list of strings, including the names to match against,
    #               iterate through all the permuations of prefixes ordering
    #               and combine them with the strings to generate a list
    #               of argument to be passed to _match_opt_strings().
    #               assert that at each iteration, the yielded tuple
    #               is the expected one.

    for prefix_combo in itertools.product(prefixes, repeat=len(raw_names)):
        arg_strs = []
        for i, p in enumerate(prefix_combo):
            name = raw_names[i]
            opt_str = f"{p}{name}"
            arg_strs.append(opt_str)

        print(arg_strs)
        idx_name_tuples = [
            (i, name) for i, name in _match_opt_strings(names, prefixes, arg_strs)
        ]
        assert idx_name_tuples == expected


@pytest.mark.parametrize(
    "prefixes, names, expected, args_in",
    [
        (
            ["-"],
            ["b", "bleh"],
            [(1, "b"), (4, "bleh")],
            ["b", "-b", "+b", "/b", "-bleh", "bleh", "+bleh", "/bleh"],
        ),
        (
            ["+"],
            ["b", "bleh"],
            [(2, "b"), (6, "bleh")],
            ["b", "-b", "+b", "/b", "-bleh", "bleh", "+bleh", "/bleh"],
        ),
        (
            ["/"],
            ["b", "bleh"],
            [(3, "b"), (7, "bleh")],
            ["b", "-b", "+b", "/b", "-bleh", "bleh", "+bleh", "/bleh"],
        ),
        (
            ["-", "+", "/"],
            ["b", "bleh"],
            [(1, "b"), (2, "b"), (3, "b"), (7, "bleh"), (9, "bleh"), (10, "bleh")],
            [
                "b",
                "-b",
                "+b",
                "/b",
                "-bbb",
                "+bbb",
                "/bbb",
                "-bleh",
                "bleh",
                "+bleh",
                "/bleh",
                "-blehelo",
                "+blehelo",
                "/blehelo",
            ],
        ),
    ],
)
def test_match_opt_strings(
    prefixes: List[str],
    names: List[str],
    expected: List[Tuple[int, str]],
    args_in: List[str],
):
    idx_name_tuples = [t for t in _match_opt_strings(names, prefixes, args_in)]
    assert idx_name_tuples == expected


@pytest.mark.parametrize(
    "prefix, alias_a, alias_b, alias_c, alias_d, exp_a, exp_b, exp_c, exp_d",
    [
        ("-", "all", "bleh", "cringe", "dude", 10, 3.142, "crack", True),
        ("+", "all", "bleh", "cringe", "dude", 10, 3.142, "crack", True),
        ("/", "all", "bleh", "cringe", "dude", 10, 3.142, "crack", True),
    ],
)
def test_alias(
    prefix: str,
    alias_a: str,
    alias_b: str,
    alias_c: str,
    alias_d: str,
    exp_a: int,
    exp_b: float,
    exp_c: str,
    exp_d: bool,
):
    @cook
    class Recipe:
        a: int = ingredient(alias=alias_a, default=None)
        b: float = ingredient(alias=alias_b, default=None)
        c: str = ingredient(alias=alias_c, default=None)
        d: bool = ingredient(alias=alias_d)

    normal_args = [
        f"{prefix}a {exp_a}",
        f"{prefix}b {exp_b}",
        f"{prefix}c {exp_c}",
        f"{prefix}d",
    ]
    alias_args = [
        f"{prefix}{alias_a} {exp_a}",
        f"{prefix}{alias_b} {exp_b}",
        f"{prefix}{alias_c} {exp_c}",
        f"{prefix}{alias_d}",
    ]

    served_norm = serve(
        Recipe, raw_ingredients=" ".join(normal_args).split(), prefixes=[prefix]
    )
    assert served_norm.a == exp_a
    assert served_norm.b == exp_b
    assert served_norm.c == exp_c
    assert served_norm.d == exp_d

    for perm in itertools.permutations(alias_args):
        served_alias = serve(
            Recipe, raw_ingredients=" ".join(perm).split(), prefixes=[prefix]
        )
        assert served_alias.a == exp_a
        assert served_alias.b == exp_b
        assert served_alias.c == exp_c
        assert served_alias.d == exp_d
        assert served_norm == served_alias

    # if alias is the same name as the ingredient name
    # should rase ValueError on serve()
    @cook
    class WrongRecipe:
        hello: str = ingredient(alias="hello", default=None)

    with pytest.raises(ValueError):
        serve(WrongRecipe, raw_ingredients=["world"])
